package scancode;

import java.io.PrintWriter;
import java.io.FileWriter;
import java.io.File;

/**
 * Copyright (C) 2010  Molly Megraw
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact: molly.megraw@duke.edu
 *
 */
public class GenerateScores {
    public GenerateScores() {
    }

    public static void main(String[] args) throws java.io.IOException {
        GenerateScores generatescores = new GenerateScores();

        double[] bg = null;
        double[][] mo = null;
        RandomNumberGenerator r = new RandomNumberGenerator();

        boolean USE_MARKOV_1_BG = (new Boolean(args[0])).booleanValue();
        double PseudoCountsVal = Double.parseDouble(args[1]);
        int N = Integer.parseInt(args[2]);
        String pwms_Fname = args[3];
        String randScores_Fname = args[4];
        String pwmScores_Fname = args[5];
        String bgFile = args[6];
        String bg2File = null;

        if (args.length > 7) { bg2File = args[7]; }

        // Report Markov Order
        System.out.println("USE_MARKOV_1_BG = " + Boolean.toString(USE_MARKOV_1_BG));

        // Load background file(s)
        bg = Load.loadVector(new File(bgFile));
        if (bg2File != null) {
            mo = Load.loadMatrix(new File(bg2File));
        }

        // Open files for printing scores for random and PWM-drawn sequences
        PrintWriter randScoresFile = new PrintWriter(new FileWriter(randScores_Fname));
        PrintWriter pwmScoresFile = new PrintWriter(new FileWriter(pwmScores_Fname));

        // Read PWM file
        PWMReturn pwms = Load.loadPWMFile(pwms_Fname, PseudoCountsVal);

        // Process each pwm from the input file
        double[][] ThetaT;
        for (int nmat = 0; nmat < pwms.pwms.length; nmat++) {

            ThetaT = pwms.pwms[nmat];

            char[][] S;
            int ind;
            double[][] Scores;

            // Generate N sequences of length defined by Theta from random background array bg,
            // score them and print the row of scores
            S = new char[N][ThetaT.length];
            for (int i = 0; i < S.length; i++) {
                for (int j = 0; j < S[i].length; j++) {
                    try {
                        ind = Prob.multinomial(bg,r);
                    }
                    catch (BadDistributionException bde) {
                        throw new IllegalArgumentException("Each column of motif frequency matrix must be a probability distribution.");
                    }
                    S[i][j] = Setup.indChar(ind, true);
                }
            }
            Scores = getScores(USE_MARKOV_1_BG, S, ThetaT, bg, mo);
            randScoresFile.print(pwms.labels[nmat]);
            for (int i = 0; i < Scores.length; i++) {
                for (int j = 0; j < Scores[i].length; j++) {
                    randScoresFile.print("\t" + Scores[i][j]);
                }
            }
            randScoresFile.println();

            // Generate N sequences from PWM ThetaT,
            // score them and print the row of scores
            S = new char[N][ThetaT.length];
            for (int i = 0; i < S.length; i++) {
                for (int mp = 0; mp < S[i].length; mp++) {
                    try {
                        ind = Prob.multinomial(ThetaT[mp], r);
                    }
                    catch (BadDistributionException bde) {
                        throw new IllegalArgumentException(
                                "Each column of motif frequency matrix must be a probability distribution.");
                    }
                    S[i][mp] = Setup.indChar(ind, true);
                }
            }
            Scores = getScores(USE_MARKOV_1_BG, S, ThetaT, bg, mo);
            pwmScoresFile.print(pwms.labels[nmat]);
            for (int i = 0; i < Scores.length; i++) {
                for (int j = 0; j < Scores[i].length; j++) {
                    pwmScoresFile.print("\t" + Scores[i][j]);
                }
            }
            pwmScoresFile.println();

            if (nmat % 10 == 0 ) {
                System.out.println("Processed pwm " + nmat);
            }

        }

        randScoresFile.close();
        pwmScoresFile.close();

    }

    public static double[][] getScores(boolean USE_MARKOV_1_BG, char[][] S, double[][] ThetaT, double[] B, double[][] B_M1) {

        int w = ThetaT.length;
        int chrInd, chrInd_prev;
        double theta_freq, bg_freq, score;
        double numLogSum, denLogSum;
        double[][] Scores;

        if (B_M1 != null) {
            USE_MARKOV_1_BG = true;
        }

        Scores = new double[S.length][];
        for (int i = 0; i < S.length; i++) {
            Scores[i] = new double[S[i].length - w + 1];
        }

        // Compute scores for each possible motif site
        for (int i = 0; i < S.length; i++) {
            for (int j = 0; j < (S[i].length - w + 1); j++) {
                numLogSum = 0.0;
                denLogSum = 0.0;
                // Take product of freq/bg probabilities over motif width
                for (int mp = 0; mp < w; mp++) {
                    chrInd = Setup.charInd(S[i][j+mp]);
                    theta_freq = ThetaT[mp][chrInd];
                    if (USE_MARKOV_1_BG) {
                        if (mp == 0) {
                            bg_freq = B[chrInd];
                        } else {
                            chrInd_prev = Setup.charInd(S[i][j + mp - 1]);
                            bg_freq = B_M1[chrInd_prev][chrInd];
                        }
                    }
                    else {
                        bg_freq = B[chrInd];
                    }
                    numLogSum += Math.log(theta_freq);
                    denLogSum += Math.log(bg_freq);
                }
                score = numLogSum - denLogSum;
                if (Double.isInfinite(score)) {
                    System.out.println("WARNING: Encountered an infinite score at position (" + i + "," + j + ")");
                }
                Scores[i][j] = score;
            }
        }
        return Scores;
    }

}
