package scancode;

/**
 * Copyright (C) 2010  Molly Megraw
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact: molly.megraw@duke.edu
 *
 */
public class Prob {

    private static double Epsilon = 1e-10;

    public Prob() {
    }

    /**
     * Returns a multinomial draw from the defined probability distribution.
     *
     * @param p       Probability distribution
     *
     * @param r       RandomNumberGenerator object
     *
     * @return        Index of object drawn
     *
     * @exception     BadDistributionException  if input vector is not a probability distribution
     *
     */
    public static int multinomial(double[] p, RandomNumberGenerator r) throws BadDistributionException {

        int n, draw;
        double u;
        double[] cdf;

        n = p.length;
        if (n < 1) {
            throw new BadDistributionException("multinomial: Not a probability distribution.");
        }

        cdf = new double[n];
        cdf[0] = p[0];
        for (int i = 1; i < n; i++) {
            cdf[i] = cdf[i - 1] + p[i];
        }
        if (!equalsOne(cdf[n-1])) {
            throw new BadDistributionException("multinomial: Not a probability distribution.");
        }

        u = r.nextUniform();
        draw = n - 1;
        for (int i = 0; i < n; i++) {
            if (u <= cdf[i]) {
                draw = i;
                break;
            }
        }

        return draw;
    }

    /**
     * Returns an array of integers from 0 to N-1 drawn randomly without replacement
     *
     * @param N       Positive integer
     *
     * @param r       RandomNumberGenerator object
     *
     * @return        Array of drawn integers
     *
     */
    public static int[] drawWithoutReplacement(int N, RandomNumberGenerator r) {

        int rnd;
        int[] numbers = new int[N];

        for (int i = 0; i < N; i++) {
            rnd = ((int)(Math.random()*N));
            numbers[i]=rnd;
            if (i > 0) {
                for (int j = i; j > 0; j--) {
                    if (numbers[i] == numbers[j-1])
                        i--;
                }
            }
        }

        return numbers;
    }

    /**
     * Tests whether a double is extremely close to 1.0.  Useful in testing
     * whether an array is a probability distribution.
     *
     * @param num  Number which is a double
     *
     * @return     True if number is extremely close to 1.0, false otherwise.
     */
    public static boolean equalsOne(double num) {

        if (num > 1.0 + Epsilon || num < 1.0 - Epsilon) {
            return false;
        }
        else {
            return true;
        }

    }

}
