package scancode;

import java.io.File;
import java.io.PrintWriter;
import java.io.FileWriter;
import java.util.Vector;
import java.util.Collections;

/**
 * Copyright (C) 2010  Molly Megraw
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact: molly.megraw@duke.edu
 *
 */
public class Scan {
    public Scan() {
     }

     public static void main(String[] args) throws java.io.IOException {
         Scan scansites = new Scan();

         // Set variable which represents the number of nucleotides after the
         // TSS represented in the FASTA file of sequences.  For example,
         // if the TSS is at the "far right" of each sequence, then set
         // nucsAfterTSS = 0.  If the TSS is 500 nucleotides from the "far right"
         // of each sequence, then set nucsAfterTSS = 500.
         int nucsAfterTSS = 0;

         char[][] S;
         PWMReturn pwms;
         DoubleColReturn minScores;
         double[] bg = null;
         double[][] bg2 = null;

         double PseudoCountsVal = Double.parseDouble(args[0]);
         String strand = args[1]; // FWD or REV - designates sense or antisense strand search
         String pwms_Fname = args[2];
         String minScores_Fname = args[3];
         String seqs_Fname = args[4];
         String out_Fbase = args[5];
         String bg_Fname = args[6];
         String bg2_Fname = null;

         if (args.length > 7) { bg2_Fname = args[7]; }

         // Load background file(s)
         bg = Load.loadVector(new File(bg_Fname));
         if (bg2_Fname != null) {
             bg2 = Load.loadMatrix(new File(bg2_Fname));
            System.out.println("Using 1st Order Markov Background");
        }

         // Read promoter sequences file
         FastaReturn Seqs = Load.loadFastaFile(new File(seqs_Fname));
         S = new char[Seqs.lines.length][];
         for (int i = 0; i < Seqs.lines.length; i++) {
             S[i] = Seqs.lines[i].toCharArray();
         }
         // Get labels for seqs
         String[] seqLabels = new String[Seqs.headers.length];
         for (int i = 0; i < seqLabels.length; i++) {
             String header = Seqs.headers[i].substring(0, Seqs.headers[i].length());
             String[] header_parts = header.split("\\s+");
             seqLabels[i] = header_parts[0];
         }

         // Read PWM file
         pwms = Load.loadPWMFile(pwms_Fname, PseudoCountsVal);

         // Read score threshold file corresponding to PWM file
         minScores = Load.loadDoubleColumnFile(new File(minScores_Fname));

         // Open file for printing count observations
         PrintWriter outFileHitsDetail = new PrintWriter(new FileWriter(out_Fbase + "." + strand + ".locations"));
         PrintWriter outFileScoresDetail = new PrintWriter(new FileWriter(out_Fbase + "." + strand + ".scores"));

         // Scan each PWM over the sequence set
         double[][] ThetaT;
         for (int nmat = 0; nmat < pwms.pwms.length; nmat++) {

             ThetaT = pwms.pwms[nmat];

             PWMHitRec hitrec = Scan(S, ThetaT, strand, bg, bg2, minScores.values[nmat], nucsAfterTSS);

             // Print hit and score distributions
             Collections.sort(hitrec.locVec);
             Collections.sort(hitrec.scoreVec);
             int nhits = hitrec.locVec.size();

             // Print hit and score detail record
             if (nhits > 0) {
                 int nSeqsHit = (int) (hitrec.proportionHit*S.length);
                 outFileHitsDetail.println(">\t" + pwms.labels[nmat] + "\t" + nSeqsHit);
                 outFileScoresDetail.println(">\t" + pwms.labels[nmat] + "\t" + nSeqsHit);
                 for (int seq = 0; seq < hitrec.hitLocs.length; seq++) {
                     if (hitrec.hitLocs[seq].length > 0) {
                         outFileHitsDetail.print(seqLabels[seq]);
                         outFileScoresDetail.print(seqLabels[seq]);
                         for (int hit = 0; hit < hitrec.hitLocs[seq].length; hit++) {
                             outFileHitsDetail.print("\t" + hitrec.hitLocs[seq][hit]);
                             outFileScoresDetail.print("\t" + Print.df3.format(hitrec.hitScores[seq][hit]));
                         }
                         outFileHitsDetail.print("\n");
                         outFileScoresDetail.print("\n");
                     }
                 }

             }

             outFileHitsDetail.flush();
             outFileScoresDetail.flush();

         }

         outFileHitsDetail.close();
         outFileScoresDetail.close();
     }

     public static PWMHitRec Scan (char[][] S, double[][] in_ThetaT, String strand, double[] B, double[][] B_M1, double minScore, int beyondTSS) {

         double[][] ThetaT;
         if (strand.equals("FWD")) {
             ThetaT = in_ThetaT;
         }
         else { // REV
             ThetaT = Utils.PWMRevCmp(in_ThetaT);
         }

         int w = ThetaT.length;
         int chrInd, chrInd_prev, hitCount, seqsHit;
         double theta_freq, bg_freq, score;
         double numLogSum, denLogSum;
         double prop;
         int[][] hitLocs = new int[S.length][];
         double[][] hitScores = new double[S.length][];
         Vector locVec = new Vector();
         Vector scoreVec = new Vector();
         boolean acceptMotif;

         boolean USE_MARKOV_1_BG = false;

         if (B_M1 != null) {
             USE_MARKOV_1_BG = true;
         }

         // Compute scores for each possible motif site
         seqsHit = 0;
         for (int i = 0; i < S.length; i++) {
             Vector hitLocsVec = new Vector();
             Vector hitScoresVec = new Vector();
             hitCount = 0;
             for (int j = 0; j < (S[i].length - w + 1); j++) {

                 // Take product of freq/bg probabilities over motif width
                 numLogSum = 0.0;
                 denLogSum = 0.0;
                 acceptMotif = true;
                 for (int mp = 0; mp < w; mp++) {
                     chrInd = Setup.charInd(S[i][j+mp]);
                     if (chrInd < 0) { // encountered an unrecognized character
                         acceptMotif = false;
                         break;
                     }
                     theta_freq = ThetaT[mp][chrInd];
                     if (USE_MARKOV_1_BG) {
                         if (mp == 0) {
                             bg_freq = B[chrInd];
                         } else {
                             chrInd_prev = Setup.charInd(S[i][j + mp - 1]);
                             bg_freq = B_M1[chrInd_prev][chrInd];
                         }
                     }
                     else {
                         bg_freq = B[chrInd];
                     }
                     numLogSum += Math.log(theta_freq);
                     denLogSum += Math.log(bg_freq);
                 }

                 // Only consider motif if all characters are recognized
                 if (acceptMotif == true) {
                     score = numLogSum - denLogSum;
                     if (score > minScore) {
                         hitCount++;
                         int pos = j;
                         int tssLoc = S[i].length - beyondTSS;
                         int hitLoc = pos - tssLoc;

                         // Record hit location and score by sequence
                         hitLocsVec.addElement(new Integer(hitLoc));
                         hitScoresVec.addElement(new Double(score));

                         // Record vectors of all hit locations and scores
                         locVec.addElement(new Integer(hitLoc));
                         scoreVec.addElement(new Double(score));

                     }
                     if (Double.isInfinite(score)) {
                         System.out.println("WARNING: Encountered an infinite score at position (" + i + "," + j + ")");
                     }
                 }

             }

             if (hitCount > 0) {
                 seqsHit++;
             }


             Object[] hitLocsArr = hitLocsVec.toArray();
             hitLocs[i] = new int[hitLocsArr.length];
             Object[] hitScoresArr = hitScoresVec.toArray();
             hitScores[i] = new double[hitScoresArr.length];
             for (int k = 0; k < hitLocsArr.length; k++) {
                 hitLocs[i][k] = ((Integer) hitLocsArr[k]).intValue();
                 hitScores[i][k] = ((Double) hitScoresArr[k]).doubleValue();
             }

         }

         prop = ((double) seqsHit)/((double) S.length);
         return new PWMHitRec(prop, hitLocs, locVec, hitScores, scoreVec);

     }

}
